library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity FrequencySense is
    generic (
        g_ClkFrequency: positive; -- input clock frequency in Hz
        g_SenseFrequencyWidth: positive -- to accomodate maximum frequency to sense
    );
    port (
        Clk_ik: in std_logic;
        SenseClk_ik: in std_logic;
        SenseFrequency_ob: out unsigned(g_SenseFrequencyWidth-1 downto 0) := (others => '0')
    );
end entity;

architecture syn of FrequencySense is

    -- SD ~ SenseClk_ik Domain

    signal Latch, LatchToggle: std_logic := '0';
    signal LatchToggleSD, LatchSD: std_logic;
    signal
        FrequencyCounterValueSD_b, FrequencyCounterValueSD_bd,
        FrequencyCounterValue_b, FrequencyCounterValue_bd:
            std_logic_vector(g_SenseFrequencyWidth-1 downto 0);
    signal Frequency_b: unsigned(g_SenseFrequencyWidth-1 downto 0);

begin

    -- Clk_ik domain

    c1HzGenerator: entity work.CounterLength(syn)
        generic map (
            g_Length => g_ClkFrequency
        )
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => '0',
            Enable_i => '1',
            Overflow_o => Latch
        );

    pLatchToggle: process (Clk_ik) is begin
        if rising_edge(Clk_ik) then
            if Latch = '1' then
                LatchToggle <= not LatchToggle;
            end if;
        end if;
    end process;
    
    cFrequencyCounterReg: entity work.Reg(syn)
        port map (
            Clk_ik => Clk_ik,
            Reset_ir => '0',
            Data_ib => FrequencyCounterValue_b,
            Enable_i => Latch,
            Data_ob => FrequencyCounterValue_bd
        );
    
    Frequency_b <= unsigned(FrequencyCounterValue_b) - unsigned(FrequencyCounterValue_bd);

    pFrequencyReg: process (Clk_ik) is begin
        if rising_edge(Clk_ik) then
            if Latch = '1' then
                SenseFrequency_ob <= Frequency_b;
            end if;
        end if;
    end process;
    
    -- CDC
    
    cLatchSyncer: entity work.Delay(syn)
        generic map (
            g_Width => 1,
            g_Delay => 3,
            g_AsyncRegUsed => "TRUE"
        )
        port map (
            Clk_ik => SenseClk_ik,
            Data_ib(0) => LatchToggle,
            Data_ob(0) => LatchToggleSD
        );
    
    cFrequencyCounterSyncer: entity work.Delay(syn)
        generic map (
            g_Width => g_SenseFrequencyWidth,
            g_Delay => 3,
            g_AsyncRegUsed => "TRUE"
        )
        port map (
            Clk_ik => Clk_ik,
            Data_ib => FrequencyCounterValueSD_bd,
            Data_ob => FrequencyCounterValue_b
        );
    
    -- SenseClk_ik domain
    
    cFrequencyCounterSD: entity work.Counter(syn)
        generic map (
            g_Width => g_SenseFrequencyWidth
        )
        port map (
            Clk_ik => SenseClk_ik,
            Reset_ir => '0',
            Enable_i => '1',
            Set_i => '0',
            SetValue_ib => (others => '0'),
            Overflow_o => open,
            std_logic_vector(Value_ob) => FrequencyCounterValueSD_b
        );
    
    cLatchEdgeDetectorSD: entity work.EdgeDetector(both)
        port map (
            Clk_ik => SenseClk_ik,
            Signal_i => LatchToggleSD,
            Edge_o => LatchSD
        );
    
    cFrequencyCounterRegSD: entity work.Reg(syn)
        port map (
            Clk_ik => SenseClk_ik,
            Reset_ir => '0',
            Data_ib => FrequencyCounterValueSD_b,
            Enable_i => LatchSD,
            Data_ob => FrequencyCounterValueSD_bd
        );

end architecture;